/**
   
 */
/*  2000 Serge Rosmorduc,
 * quipe EC.art, Laboratoire LRIA, Universit de Paris 8
 *
 * 15 rue Catulienne
 * 93236 St Denis Cedex
 * FRANCE
 *
 * Tout droits rservs.
 * 
 * voyez le fichier ../LICENCE.txt pour plus d'informations
 * See the ../LICENCEE.txt file for more information
 *
 * $Id$
 *
 */

/*
 * $Log$
 */

#ifndef EXPR_H
#define EXPR_H

// In theory, everything will be in the Transductor namespace.

#include <string>
#include <list>

using namespace std;		/* NO-OP in GCC, included for portability */

class Expr;
class PlainExpr;
class VarExpr;
class FuncExpr;
class NegExpr;
class BoundExpr;
class EpsilonExpr;

#include "Bindings.h"
#include "PointerList.h"

class PlainExpr {
public:
  PlainExpr();
  virtual ~PlainExpr();
  // The match method sends back a pointer to 
  // Bindings compatible with the match, or 0 if no such binding exist.
  // v : expression to match
  // b1 : bindings for this
  // b2 : bindings for v
  virtual bool match(const PlainExpr *v, Bindings *b1, Bindings *b2) const = 0 ;
  virtual bool matchVar(const VarExpr *v, Bindings *b1, Bindings *b2) const = 0;
  virtual bool matchNeg(const NegExpr *v, Bindings *b1, Bindings *b2) const =0;
  virtual bool matchFunc(const FuncExpr *v, Bindings *b1, Bindings *b2) const = 0;
  virtual string toString(const Bindings *b) const=0 ;
  virtual bool isEpsilon() const {return false;};
 private:
  // Bindings * nextBindings;
};

class VarExpr : public PlainExpr {
 public:
  // VarExpr(const string& s);
  VarExpr(const VarRep *);
  ~VarExpr();
  bool match(const PlainExpr *v, Bindings *b1, Bindings *b2) const;
  bool matchVar(const VarExpr *v, Bindings *b1, Bindings *b2) const;
  bool matchNeg(const NegExpr *e, Bindings *b1, Bindings *b2) const;
  bool matchFunc(const FuncExpr *v, Bindings *b1, Bindings *b2) const;
  bool hasvalue(const Bindings *b1) const;
  string toString(const Bindings *b) const;
  void setVar(const VarRep *v);
  const VarRep *getVar() const;
  /* Sets the value of this expression in b1 as v. */
  void setVal(const BoundExpr *v, Bindings *b1) const;
  const BoundExpr *getVal(const Bindings *b1) const;
 private:
  const VarRep *var;
};


class FuncExpr : public PlainExpr {
 public:
  FuncExpr(const string&  f, list<PlainExpr *> *e);
  FuncExpr(const string&  f);
  ~FuncExpr();
  // FuncExpr(const FuncExpr& e);
  bool match(const PlainExpr *v, Bindings *b1, Bindings *b2) const;
  bool matchVar(const VarExpr *v, Bindings *b1, Bindings *b2) const;
  bool matchNeg(const NegExpr *e, Bindings *b1, Bindings *b2) const;
  bool matchFunc(const FuncExpr *v, Bindings *b1, Bindings *b2) const;
  string  toString(const Bindings *b) const;
  const string& getFunctor() const;
  void setFunctor(string s);
  list<PlainExpr *> const *getArgs() const;
  void setArgs(list<PlainExpr *> *args);
 private:
  string functor;
  PointerList<PlainExpr> arg;
};

class NegExpr : public PlainExpr {
public:
  NegExpr(const FuncExpr *exp);
  ~NegExpr();
  bool match(const PlainExpr *v, Bindings *b1, Bindings *b2) const;
  bool matchVar(const VarExpr *v, Bindings *b1, Bindings *b2) const;
  bool matchNeg(const NegExpr *v, Bindings *b1, Bindings *b2) const;
  bool matchFunc(const FuncExpr *v, Bindings *b1, Bindings *b2) const;
  string toString(const Bindings *b) const;
  const FuncExpr * getExpr() const;
 private:
  const FuncExpr *exp; // Negative expression cannot be variable, even after instantiation
};

class EpsilonExpr : public PlainExpr {
 public:
  EpsilonExpr();
  ~EpsilonExpr();
  bool match(const PlainExpr *v, Bindings *b1, Bindings *b2) const;
  bool matchVar(const VarExpr *v, Bindings *b1, Bindings *b2) const;
  bool matchNeg(const NegExpr *v, Bindings *b1, Bindings *b2) const;
  bool matchFunc(const FuncExpr *v, Bindings *b1, Bindings *b2) const;
  string toString(const Bindings *b) const;
  bool isEpsilon() const {return true;};
};

class AndExpr {
 public:
  AndExpr(list<PlainExpr *> *terms);
  ~AndExpr();
  string toString(const Bindings *b) const;
  bool match(const PlainExpr *e, Bindings *b1, Bindings *b2) const;
  const list<PlainExpr *> * getTerms() const;
 private:
  PointerList<PlainExpr> terms;
};

class Expr {
 public:
  Expr(AndExpr *a);
  ~Expr();
  string toString(const Bindings *b) const;
  bool match(const PlainExpr *e, Bindings *b1, Bindings *b2) const;
  bool isEpsilon() const;

  static const Expr * getEpsilon();
 private:
  Expr();
  AndExpr *rep;
  Expr(const Expr &e) {throw new string("Illegal");};

  // Constant for the class
  static const Expr * Expr::epsilonExpr;
};

/**
 * BoundedExpr is an expression plus its bindings
 */

class BoundExpr {
 public:
  BoundExpr(const PlainExpr *e, const Bindings *b) {bindings= b; expr= e;};
  const PlainExpr * getExpr() const {return expr;};
  const Bindings *getBindings() const {return bindings;};
  string toString() const;
  
  /**
   * Matches the bound expression with the unbound expression v.
   * new bindings for v go in newBindings.
   */
  bool match(const PlainExpr *v, Bindings *newBindings) const;
  static const BoundExpr * getEpsilon();
 private:
  const PlainExpr *expr;
  const Bindings *bindings;

  // Constant for the class
  static const BoundExpr * epsilonExpr;

};

ostream & operator<<(ostream &s, const BoundExpr &e);

#endif

