/**
   
 */
/*  2000 Serge Rosmorduc,
 * quipe EC.art, Laboratoire LRIA, Universit de Paris 8
 *
 * 15 rue Catulienne
 * 93236 St Denis Cedex
 * FRANCE
 *
 * Tout droits rservs.
 * 
 * voyez le fichier ../LICENCE.txt pour plus d'informations
 * See the ../LICENCEE.txt file for more information
 *
 * $Id$
 *
 */

/*
 * $Log$
 */


// Grammar : 
// Rule  ::= LEFT "->" RIGHT [ "/" COST ]
// LEFT  ::= ANDEXPR | ANDEXPR "," LEFT
// ANDEXPR ::= ITEM | ITEM "and" ANDEXPR
// RIGHT ::= ITEM | ITEM "," RIGHT
// ITEM  ::= EPSILON | VAR | FUNC
// FUNC  ::= ID | ID "(" ATOMS ")"
// ATOMS ::= ATOM | ATOM "," ATOMS
// ATOM ::= VAR | ID

// Parser for an Expression.

// Conventions :
// 	The first token has always been read and can be consulted.
//	So it's the method's responsability to read the token which follows 
//	its end (in other terms, a method should more or less end with t->next())
//
//	The test for the first token's value will depend on the grammar.
//	we will put it in the most practical place, which might be the caller or
//	the callee, depending on the rule.

// Issues : 
//   Should the parser be the classes' constructor or not ?
//   Pro:
//	The way the classes are written, the best place for deleting
//	any part of a rule is the destructor.
//	If the destructor is responsible for freeing something, the 
//	constructor should be the one in charge of building it, for
//	symetry's sake.
//
//   Cons:
//	There's no compeling reason why a rule should know about its
//	external representation. For instance, the tokeniser should
//	not leave longer than the rule construction. As all parts of
//	rule construction must know about it, it's logical for it to
//	be a member part of a "builder" object.  
//
//   Best solution would be :
//	The rule constructor builds up a builder object.
//
//   Alternative solution is :
//	The only way to build an Rule is by using the builder =>
//	make the Rule constructor private and make the builder class a
//	friend of rule. 
//
// The way the stuff is done now, SimpleExprBuilder could be part of
// the Rule class.  If we think about it, Rule can not exist without
// SimpleExprBuilder (it's the only way to build a Rule).


#include "ExprBuilder.h"

AndExpr *
SimpleExprBuilder::parseAndExpr()
{
  list<PlainExpr *> *tmp= new list<PlainExpr*>;
  try {
    tmp->push_back(parseItem());
    while (t->getToken() == ExprTokeniser::AND)
      {
	t->next();
	tmp->push_back(parseLeftItem());
      }
  }
  catch (SyntaxError &e) {
    delete tmp;
    throw;
  }
  return new AndExpr(tmp);
}

PlainExpr *
SimpleExprBuilder::parseAtom()
{
  PlainExpr * res= 0;
  if (t->getToken() == ExprTokeniser::ID
      || t->getToken() == ExprTokeniser::NUM)
    {
      res= new FuncExpr(t->getTokenValue());
      t->next();
    }
  else if (t->getToken() == ExprTokeniser::VAR)
    {
      VarRep * v= varTable->getVar(t->getTokenValue());
      res= new VarExpr(v);
      t->next();
    }
  else
    {
      SyntaxError err("Unexpected token after : " );
      throw err;
    }
  return res;
}

list<PlainExpr *> * 
SimpleExprBuilder::parseArgs()
{
  t->next(); // Ok, we know for sure first token is a '('
  list<PlainExpr *> * res= new list<PlainExpr *>();
  try{
    PlainExpr * first= parseAtom();
    res->push_back(first);
    while (t->getToken() == ',')
    {
      t->next();
      PlainExpr * expr= parseAtom();
      res->push_back(expr);
    }
    t->next();			// read past the ')'
  }
  catch (SyntaxError &e) {
    delete res;
    throw;
  }
  return res;
}

FuncExpr *
SimpleExprBuilder::parseFunc()
{
  FuncExpr* res= 0;
  string debut= t->getTokenValue();
  t->next();
  if (t->getToken() == '(')
    {
      list<PlainExpr*> * args= parseArgs();
      res= new FuncExpr(debut, args);
    }
  else
    {
      res= new FuncExpr(debut);
    }
  return res;
}

VarExpr *
SimpleExprBuilder::parseVar()
{
  VarRep * v= varTable->getVar(t->getTokenValue());
  VarExpr * res= new VarExpr(v);
  t->next();
  return res;
}

PlainExpr *
SimpleExprBuilder::parseItem()
{
  PlainExpr* res= 0; 
  // Item should start either with a ID or with a VAR.
  if (t->getToken() == ExprTokeniser::ID 
      || t->getToken() == ExprTokeniser::NUM)
    {
      // It's a FUNC
      res= parseFunc();
    }
  else if (t->getToken() == ExprTokeniser::VAR)
    {
      // It's a VAR
      res= parseVar();
    }
  else if (t->getToken() == ExprTokeniser::EPSILON)
    {
      res= new EpsilonExpr();
      t->next();
    }
  else 
    {
      // It's an error.
      throw SyntaxError("Unexpected token " + t->getTokenValue());
    }
  return res;
}

PlainExpr *
SimpleExprBuilder::parseLeftItem()
{
  if (t->getToken() == ExprTokeniser::NEG)
    {
      t->next();
      return new NegExpr(parseFunc());
    }
  else
    return parseItem();
}

list<PlainExpr *> *
SimpleExprBuilder::parseTerms()
{
  list<PlainExpr*> *res= new list<PlainExpr*>; 
  // CHANGE IT TO :
  /*   if (t->getToken() == ExprTokeniser::EPSILON)
    {
      left= 0;
      t->next();
    }
    and suppress EPSILON in parseItem.
    EPSILON means "empty list", that's all.
  */
  res->push_back(parseItem());
  while (t->getToken() == ',')
    {
      t->next();
      res->push_back(parseItem());
    }
  return res;
}

list<Expr *> *
SimpleExprBuilder::parseItems()
{
  list<Expr*> *res= new list<Expr*>; 
  if (t->getToken() == ExprTokeniser::EPSILON)
    {
      t->next();
    }
  else
    {
      res->push_back(new Expr(parseAndExpr()));
      while (t->getToken() == ',')
	{
	  t->next();
	  res->push_back(new Expr(parseAndExpr()));
	}
    }
  return res;
}

Rule *
SimpleExprBuilder::parseRule()
{  
  list<Expr *> *left= 0;
  list<PlainExpr *> *right= 0;
  int cost= 0;
  currentRule= new Rule();
  varTable= currentRule->getVarTable();
  try {
  t->next();
  left= parseItems();
  if (t->getToken() != ExprTokeniser::RULESEP)
    {
      throw SyntaxError("Unexpected token " + t->getTokenValue() + "\n" +
			"I was expecting  -> ");
    }
  t->next();
  right= parseTerms();
  if (t->getToken() == '/')
    {
      t->next();
      if (t->getToken() == ExprTokeniser::NUM)
	{
	  cost= t->getTokenNumValue();
	}
      else
	{
	  throw SyntaxError("Unexpected token " + t->getTokenValue() + "\n" +
			    "I was expecting  an integer cost");
	}
    }
  } 
  catch (SyntaxError &e) {
    delete left;
    delete right;
    delete currentRule;
    throw;
  }
  currentRule->setLeft(left);
  currentRule->setRight(right);
  currentRule->setCost(cost);
  return currentRule;
}

Rule * SimpleExprBuilder::makeRule(string s)
{
  Rule *res;
  t= new ExprStringTokeniser(s);
  res= parseRule();
  delete t;
  return res;
}

Rule * SimpleExprBuilder::makeRule(istream &s)
{
  Rule *res;
  t= new ExprStreamTokeniser(s);
  res= parseRule();
  delete t;
  return res;
}

PlainExpr *
SimpleExprBuilder::makePlainExpr(string s)
{
  PlainExpr *res;
  t= new ExprStringTokeniser(s);
  t->next();
  res= parseItem();
  delete t;
  return res;
}

PlainExpr * 
SimpleExprBuilder::makePlainExpr(istream &s)
{
  PlainExpr *res;
  t= new ExprStreamTokeniser(s);
  res= parseItem();
  delete t;
  return res;
}

SimpleExprBuilder::SimpleExprBuilder()
{
  t= 0;
  bindings=0;
  currentRule=0;
  varTable=0;
}

