
/**
   
 */
/*  2000 Serge Rosmorduc,
 * quipe EC.art, Laboratoire LRIA, Universit de Paris 8
 *
 * 15 rue Catulienne
 * 93236 St Denis Cedex
 * FRANCE
 *
 * Tout droits rservs.
 * 
 * voyez le fichier ../LICENCE.txt pour plus d'informations
 * See the ../LICENCEE.txt file for more information
 *
 * $Id$
 *
 */

/*
 * $Log$
 */


#include "Rule.h"
#include "StringUtils.h"
#include <iostream>
using namespace std;

// A Rule like A,B->C will be transformed into a transducer :
//
//      A          B       epsilon
// *--------->*--------->*--------->*
//   epsilon    epsilon       C
//
// The cost will be 0 for the first two arcs, and the cost of the rule
// (or 0) for the last one.
//

Rule::Rule()
{
  left= 0;
  right= 0;
  cost= 0;
  varNames= new VarTable();
  // Computed variables
  numLeft= 0;
  numRight= 0;
}

// We have a BIG conceptual problem here. Who should delete the rule's
// parts ?  The rule does not create them (the ExprBuilder does). So
// it shouldn't really be the rule's business. Yet the builder should
// not outlive its use, methink. 

Rule::~Rule()
{
  delete varNames;
}

const list<Expr *> * 
Rule::getLeft() const
{
  return left.getList();
}


const list<PlainExpr *> * 
Rule::getRight() const
{
  return right.getList();
}

int 
Rule::getCost() const
{
  return cost;
}

VarTable *
Rule::getVarTable()
{
  return varNames;
}

const VarTable *
Rule::getVarTable() const
{
  return varNames;
}

Bindings *
Rule::makeBindings() const
{
  return new Bindings(getVarTable());
}

void Rule::setLeft(list<Expr *> *l)
{
  left.setList(l);
  numLeft= l->size();
}

void Rule::setRight(list<PlainExpr *> *r)
{
  right.setList(r);
  numRight= r->size();
}

void Rule::setCost(int c)
{
  cost =c;
}

string 
Rule::toString(Bindings *b) const
{
  string res;
  list<Expr *>::const_iterator i;
  if (startsWithEpsilon())
    res="epsilon";
  else 
    {
      for(i= getLeft()->begin(); i != getLeft()->end(); i++)
	res+= (*i)->toString(b) + ", ";
      res.erase(res.length() - 2, 2);
    }
  res+= " -> ";
  list<PlainExpr *>::const_iterator j;
  if (getRight()->size() != 0)
    {
      for(j= getRight()->begin(); j != getRight()->end(); j++)
	res+= (*j)->toString(b)+ ", ";
      res.erase(res.length() - 2, 2);
    }
  res+= " / ";
  res+= intToString(getCost());
  return res;
}

bool
Rule::startsWithEpsilon() const
{
  //return  (getNumLeft() == 0);
  return  (numLeft == 0);  
}

int
Rule::getNumLeft() const
{
  return  numLeft;
}

int
Rule::getNumRight() const
{
  return  numRight;
}


/*
 * Note about RuleItems
 *
 * 
 * When requiring a ruleItem Upper and Lower part,
 * there should be an overlap :
 *
 *
 * for instance for a rule A B C -> D E F :
 *      A, B, C, 0, 0
 *      0, 0, D, E, F
 * part l  l  m  r  r
 *
 * (0 means epsilon)
 *
 * Now, there are special cases :
 *
 * a) 0 -> D => 
 *    part = middle
 *    
 * b) D -> 0 =>
 * c) 0 -> 0
 */
 
RuleItem::RuleItem(const Rule * r)
{
  rule= r;
  lpos= rule->left.begin();
  rpos= rule->right.begin();
  itemNumber= 1;
  bindings= rule->makeBindings();
  if (r->getNumLeft() > 1)
    part= left;
  else if (r->getNumLeft() == 1)
    part= middle;
  else
    part= right;
  fixLow();
  isFirst= true;
}

RuleItem::RuleItem(const RuleItem &ri)
{
  rule= ri.rule;
  lpos= ri.lpos;
  rpos= ri.rpos;
  itemNumber= ri.itemNumber;
  bindings= new Bindings(*(ri.bindings));
  part= ri.part;
  fixLow();
  isFirst= ri.isFirst;
}

RuleItem &
RuleItem::operator=(const RuleItem &ri)
{
  if ((&ri == this))
    return *this;
  delete bindings;
  rule= ri.rule;
  lpos= ri.lpos;
  rpos= ri.rpos;
  itemNumber= ri.itemNumber;
  part= ri.part;
  bindings= new Bindings(*(ri.bindings));
  if (low)
    delete low;
  fixLow();
  return *this;
}

RuleItem::~RuleItem()
{
  delete bindings;
  delete low;
}

void
RuleItem::fixLow()
{
  if (part != left)
      low= new BoundExpr(*rpos, bindings);
  else 
    low=0;
}

void
RuleItem::next()
{
  isFirst= false;
  switch (part) {
  case left:
    lpos++;
    itemNumber++;
    if (itemNumber == rule->getNumLeft())
      part= middle;
    break;
  case right:
    itemNumber++;
    rpos++;
    break;
  case middle:
    // The first item of the right part is used by the middle element
    // as its lower part. Hence, middle should fix itemNumber at 2,
    // not one.
    part= right;
    itemNumber= 2;
    rpos++;
    break;
  }
  delete low;
  fixLow();
}

RuleItem *
RuleItem::makeMatch(const BoundExpr *exp)
{
  RuleItem *res= new RuleItem(*this);

  // We have a problem at next line with C++ ideas about const : We
  // know for sure that if exp is really bounded, its bindings won't
  // change. I really don't feel like doubling the number of function
  // I write just for this.

  if (res->getUpper()->match(exp->getExpr(), 
			     res->bindings, 
			     const_cast<Bindings *>(exp->getBindings())))
    {
      extern int debuglev;
      if (debuglev > 3)
	cout << *res << " matches " << exp->getExpr()->toString(exp->getBindings()) << endl;
      // BADDDDDDDD : res->next();
    }
  else
    {
      delete res;
      res= 0;
    }
  return res;
}

/*
  RuleItem::operator bool()
  {
  return ! (part == right && (rpos == rule->right.end()));
  }
*/

const Expr *
RuleItem::getUpper() const 
{
  if (part != right)
    return *lpos;
  else
    return Expr::getEpsilon();
}

BoundExpr *
RuleItem::getLower() const 
{
  return low;
}

// The cost of a rule is currently set on the first RuleItem
// associated with it.

int 
RuleItem::getCost() const 
{
  if (isFirst)
    return rule->getCost();
  else
    return 0;
}

Bindings * 
RuleItem::getBindings()
{
  return bindings;
}

// Returns true if the ruleItem is the last part of the rule
bool
RuleItem::isFinal()
{
  bool res= false;
  switch (part) {
  case left:
    res= false;
    break;
  case middle:
    res= (rule->getNumRight() < 2);
    break;
  case right:
    res= (rule->getNumRight() == itemNumber);
    break;
  default:
    throw string("This shouldn't be!!!!!!");
  }
  return res;
}

const Rule*
RuleItem::getRule() const
{
  return rule;
}

ostream & 
operator<<(ostream &s, const RuleItem &r)
{
  s 
    << "Rule Item " 
    << r.getRule()->toString(const_cast<RuleItem&>(r).getBindings())
    << " pos ";
  switch(r.part)
    {
    case RuleItem::left: cout << "left "; break;
    case RuleItem::right: cout << "right "; break;
    case RuleItem::middle: cout << "middle "; break;
    }
  if (r.part != RuleItem::middle)
    cout << r.itemNumber;
  return cout;
}

