/**
   Classes for working with valued transducers
 */

/*  2000 Serge Rosmorduc,
 * quipe EC.art, Laboratoire LRIA, Universit de Paris 8
 *
 * 15 rue Catulienne
 * 93236 St Denis Cedex
 * FRANCE
 *
 * Tout droits rservs.
 * 
 * voyez le fichier ../LICENCE.txt pour plus d'informations
 * See the ../LICENCEE.txt file for more information
 *
 * $Id$
 *
 */

/*
 * $Log$
 */


// Forward declarations
#include "Expr.h"
#include <stack>
#include "RuleTransducer.h"

// Direction of links
enum LinkDir {forward, backward};

// Transducers
class Transducer;
class EntryTransducer;
class CompositeTransducer;

// Generic Node class
class Node;
class CompositeNode;
class EntryNode;

// Generic Link class
class Link;
class EntryLink;
// Links related to CompositeTransducer
class CompositeLink;
class ChoiceNode;
class MiddleNode;

class LinkList;

// Base class for transducers

class Transducer {
 public:
  Transducer();
  virtual ~Transducer();
  virtual Node * getStart()= 0;
  virtual const Node * getStart() const= 0;
  virtual Node * getEnd()= 0;
  virtual Node * getCurrentNode()= 0;
  virtual void nextNode()= 0;
  virtual void reset()= 0;
  virtual bool lastNode()= 0;
  // The following method only works on
  // acyclic graphs
  void computeShortestPath();
  void showPath(ostream &o, int detail=0) const;
 private:
  list<Node *> sortedNodes;
};

// The entry transducer
class EntryTransducer : public Transducer {
 public:
  EntryTransducer();
  virtual ~EntryTransducer();

  // Overloaded Methods
  Node * getStart();
  const Node * getStart() const;
  Node * getEnd();
  Node * getCurrentNode();
  void nextNode();
  void reset();
  bool lastNode();
  // "Real" access functions :
  virtual EntryNode * getTrueStart();
  virtual void setEnd(EntryNode *);
  
  virtual EntryNode * getTrueCurrentNode();  
  // Specific Methods
  EntryNode * makeNodeByLabel(int l);
  
 private:
  EntryNode *start;
  EntryNode *end;
  
  map<int,EntryNode*> nodes;
  map<int,EntryNode*>::iterator currentNode;
};

// A Composite transducer
class CompositeTransducer : public Transducer {
 public:
  CompositeTransducer(Transducer*, RuleTransducer*);
  virtual ~CompositeTransducer();
  // Overloaded Methods
  Node * getStart();
  const Node * getStart() const;
  Node * getEnd();
  Node * getCurrentNode();
  void nextNode();
  void reset();
  bool lastNode();

  // "Real" access functions :
  virtual CompositeNode * getTrueStart();
  virtual CompositeNode * getTrueCurrentNode();  

  // Specific Methods
  ChoiceNode  * getChoiceNode(Node *n, RuleNode *r);
  int nextLabel();
  int getMaxLabel();
  void addToAgenda(CompositeNode *);

  void setEnd(CompositeNode *);

 private:
  int maxLabel; // Label to use for new nodes
  
  // The table below is used to associate a Node with a couple of Nodes
  // from the two parent transducers
  typedef map<pair<Node*,RuleNode*>,ChoiceNode*> Labels;
  Labels labelValues;
  
  // Nodes still to be expanded.
  // Shouldn't be there, methink.
  stack<CompositeNode*> toDo;

  // First Node
  CompositeNode *start;
  CompositeNode *end;

  // Nodes
  map<int,CompositeNode*> nodes;
  map<int,CompositeNode*>::iterator currentNode;
  
  // Friends
  friend class CompositeNode;
};

// Base Node Class
class Node {
 public:
  Node();
  virtual ~Node();
  virtual bool isFinal() = 0;
  virtual Transducer * getTransducer()=0;
  virtual int getLabel()=0;
  virtual LinkList * makeLinks()=0;
  virtual LinkList * makePreviousLinks()=0;
 private:
  // Data used for lowest cost path calculation
  enum State {white, gray, black} state;
  Link * bestLink;
  int bestCost;
  // Auxiliary function for lowest cost path calculation
  void depthFirst(list<Node*> &l);
  // Not that good. But C++'s model of visibility
  // is not that good, either (In Eiffel, we would export depthFirst
  // to Transducer)
  friend class Transducer;
};

// A plain Node in the entry transducer

class EntryNode : public Node {
 public:
  EntryNode(EntryTransducer *a, int lab=0);
  ~EntryNode();

  // Inherited functions
  bool isFinal();
  Transducer * getTransducer();
  int getLabel();
  LinkList * makeLinks();
  LinkList * makePreviousLinks();

  // Specific functions
  void setFinal(bool);
  void setLabel(int label);

  EntryLink * addLink(EntryNode *end, 
		      const string & Up,
		      const string & down, 
		      int cost=0);
 private:
  EntryTransducer *automaton;
  int label;
  bool final;

 protected:
  list<EntryLink*> links;
  list<EntryLink*> previousLinks;
  // Friends and the like  
  friend class LinkListFromEntry;
  friend class PreviousLinkListFromEntry;
};

// Nodes in Composite transducers
// A node in a composite transducer is either
// associated with an couple of nodes (ChoiceNode)
// or with an entry node and a rule state (MiddleNode)

class CompositeNode : public Node {
 public:
  // Inherited functions
  Transducer * getTransducer();
  int getLabel();
  LinkList * makeLinks();
  LinkList * makePreviousLinks();
  
  // Specific functions
  Node *getUpNode();
  CompositeTransducer *getCompAutomaton();
  virtual void extends()= 0;
  virtual bool isDeadNode()= 0;
  CompositeLink * addLink(CompositeNode *end,
			  const BoundExpr *,	/* upper label */
			  const RuleItem *, /* Rule item. determine lower label */
			  int);	/* cost of the link */
  // OLD :  RuleLink *); /* RuleLink associated with this CompositeLink */
  virtual ~CompositeNode();
 protected:
  CompositeNode(CompositeTransducer *a, Node *e);
 private:
  Node *entry;
  int label;
  CompositeTransducer *automaton;
  // For use by CompositeLink only :
  void registerLink(CompositeLink *, enum LinkDir);
  void unRegisterLink(CompositeLink *, enum LinkDir);
  friend class MiddleNode; // TEMPORARY
 protected:
  list<CompositeLink*> links;
  list<CompositeLink*> previousLinks;
  // Friends
  friend class LinkListFromComposite;
  friend class PreviousLinkListFromComposite;
  // CompositeLink should have access to registerLink and unregisterlink
  friend class CompositeLink;
};

// Node corresponding to a choice point between different rules
class ChoiceNode : public CompositeNode {
 public:
  ChoiceNode(CompositeTransducer *a, Node *e, RuleNode *r);
  bool isFinal();
  virtual ~ChoiceNode();
  RuleNode * getRuleNode();
 private:
  void extends();
  bool isDeadNode();
  RuleNode * ruleNode;
  bool final;
};

// Intermediary nodes. The only choice stands between different entries
// A middle node can only have one entry
class MiddleNode : public CompositeNode {
 public:
  MiddleNode(CompositeTransducer *a, // the transducer
	     Node * e,		// the entry node
	     RuleItem *r,	// the rule state
	     RuleLink *rl);	// the associated rule link
  virtual ~MiddleNode();
  bool isFinal();
 private:
  void extends();
  bool isDeadNode();
  RuleItem * getRuleItem();
  RuleLink * getRuleLink();
  RuleItem * ruleItem;
  RuleLink * ruleLink;
};

// Links in general
class Link {
 public:
  Link();
  virtual ~Link();
  virtual Node * getStart()= 0;
  virtual Node * getEnd()= 0;
  virtual const BoundExpr *getUp() const= 0;
  virtual const BoundExpr *getDown() const= 0;
  virtual bool downIsEpsilon() const= 0;
  virtual int getCost() const = 0;
};


class EntryLink : public Link {
 public:
  EntryLink(EntryNode *n1, EntryNode* n2, 
	    BoundExpr *u, BoundExpr *d, int c= 0);
  // Inherited Methods
  Node * getStart();
  Node * getEnd();
  const BoundExpr *getUp() const;
  const BoundExpr *getDown() const;
  bool downIsEpsilon() const;
  int getCost() const;
  // Specific Methods
 private:
  EntryNode *start, *end;
  BoundExpr *up;
  BoundExpr *down;
  int cost;
};

// Links in the Composite transducer

// There's a problem with having both a RuleItem and a RuleLink,
// as there's a redundancy there. The best solution would be to encapsulate the two
// in a GraphRuleItem, or something like that.

class CompositeLink : public Link {
 public:
  CompositeLink(CompositeNode *, /* starting node */
		CompositeNode*, /* end node */
		const BoundExpr *,	/* upper label */
		const RuleItem *,
		int c= 0);
  ~CompositeLink();
  // Inherited Methods
  Node * getStart();
  Node * getEnd();
  const BoundExpr *getUp() const;
  const BoundExpr *getDown() const;
  bool downIsEpsilon() const;
  int getCost() const;
  // Specific Methods
  // RuleLink * getRuleLink();
  // void setEnd (CompositeNode *);
 private:
  CompositeNode *start, *end;
  const BoundExpr *up;
  const RuleItem * down;
  int cost;
  // Not really nice. we should probably provide a set of
  // SPECIFIC access functions
  friend bool MiddleNode::isDeadNode();
};

// List of links from a node :

class LinkList 
{ 
 public:
  virtual ~LinkList();
  virtual bool hasElements()= 0;
  virtual Link *getElement()= 0;
  virtual void next()= 0;
};

// We should use templates down there.
class LinkListFromEntry : public LinkList {
 public:
  LinkListFromEntry(EntryNode *n, LinkDir dir);
  virtual ~LinkListFromEntry();
  virtual bool hasElements();
  virtual Link *getElement();
  virtual void next();
 private:
  list<EntryLink*>::iterator rep;
  list<EntryLink*>::iterator end;
};

class LinkListFromComposite : public LinkList {
 public:
  LinkListFromComposite(CompositeNode *n, LinkDir  dir);
  virtual ~LinkListFromComposite();
  virtual bool hasElements();
  virtual Link *getElement();
  virtual void next();
 private:
  list<CompositeLink*>::iterator rep;
  list<CompositeLink*>::iterator end;
};

ostream & operator<<(ostream &, Transducer &);
ostream & operator<<(ostream &, Link &);
ostream & operator<<(ostream &, Node &);
